(* Title:  MathEngBasic/problem.sml
   Author: Walther Neuper 2019
   (c) due to copyright terms

\<open>ML_structure Problem\<close> holds the essential information required for automated doing
\<open>ML_structure Specification\<close> and \<open>ML_structure Solution\<close>.
*)

signature PROBLEM =
sig
  type T = Probl_Def.T
  val empty: T

  type id = Probl_Def.id
  type id_reverse                                                  
  val id_empty: id
  val id_to_string: id -> string

  val prep_store: theory -> Check_Unique.id -> string list -> References_Def.id ->
    (Model_Pattern.m_field * (Model_Def.descriptor * term)) list * term list * term option ->
      id * References_Def.id list * Rule_Set.T -> T * id

  val set_data: Rule_Def.rule_set -> theory -> theory
  val the_data: theory -> Rule_Def.rule_set

  val parse_item: string parser -> 'a parser -> 'a parser
  val parse_item_name: string parser -> string -> string parser
  val parse_item_names: string parser -> string list parser
  val parse_pos_model_input: Model_Pattern.model_input_pos parser
  val parse_authors: string list parser
  val parse_pos_cas: (string * Position.T) option parser
  val parse_methods: string list parser
  val ml: string -> ML_Lex.token Antiquote.antiquote list

  val from_store: Proof.context -> id -> T
end

(**)
structure Problem(**): PROBLEM(**) =
struct
(**)

type T = Probl_Def.T;
val empty = Probl_Def.empty

(*
  elements if the id are in reverse order as compared to MethodC:
  e.g. ["linear", "univariate", "equation"] has "equation" as parent node --
  -- this makes the id readable.
*)
type id = Probl_Def.id;
(* same order as for MethodC (used in Store )*)
type id_reverse = Probl_Def.id;

val id_empty = Probl_Def.id_empty;
val id_to_string = Probl_Def.id_to_string;


(** prepare Problem for Know_Store **)

fun prep_store thy name maa start_refine (model, where_, cas) (pbl_id, met_ids, where_rls) =
  ({guh = name, mathauthors = maa, start_refine = start_refine, thy = thy, cas = cas,
		  where_rls = where_rls, where_ = where_, model = model, solve_mets = met_ids}, pbl_id)


(** Isar command **)

(* theory data *)

structure Data = Theory_Data
(
  type T = Rule_Set.T option;
  val empty = NONE;
  fun merge _ = NONE;
);

val set_data = Data.put o SOME;
val the_data = the o Data.get;


(* auxiliary parsers *)

fun parse_item (keyword: string parser) (parse: 'a parser) =
  (keyword -- Args.colon) |-- Parse.!!! parse;

fun parse_item_name keyword =
  Scan.optional (parse_item keyword Parse.name);

fun parse_item_names (keyword: string parser) =
  Scan.optional (parse_item keyword (Scan.repeat1 Parse.name)) [];

(*val parse_pos_tagged_terms: string parser -> string -> Token.T list ->
  (string * (string * Position.T) list) list * Token.T list*)
fun parse_pos_tagged_terms keyword (tag: string) =
  Scan.optional
    (parse_item keyword (Scan.repeat (Parse.position Parse.term)) >>
      (fn ts => [(tag, ts)])) [];

val parse_pos_model_input : Model_Pattern.model_input_pos parser =
  parse_pos_tagged_terms \<^keyword>\<open>Given\<close> "#Given" @@@
  parse_pos_tagged_terms \<^keyword>\<open>Where\<close> "#Where" @@@
  parse_pos_tagged_terms \<^keyword>\<open>Find\<close> "#Find" @@@
  parse_pos_tagged_terms \<^keyword>\<open>Relate\<close> "#Relate";

val parse_authors = parse_item_names \<^keyword>\<open>Author\<close>;
(*l parse_pos_cas: Token.T list -> (string * Position.T) option * Token.T list*)
val parse_pos_cas = Scan.option (parse_item \<^keyword>\<open>CAS\<close> (Parse.position Parse.term));
val parse_methods = parse_item_names \<^keyword>\<open>Method_Ref\<close>;
val ml = ML_Lex.read;


(* command definition *)

local

val drop_pos_model = (fn (a, b, _) => (a, b))
val drop_pos_where_ = (fn (b, _) => b)

val _ =
  Outer_Syntax.command \<^command_keyword>\<open>problem\<close>
    "prepare ISAC problem type and register it to Knowledge Store"
    (Parse.name -- Parse.!!! (\<^keyword>\<open>:\<close> |-- Parse.name --
      Parse.!!! (\<^keyword>\<open>=\<close> |-- Parse.ML_source -- parse_authors --
        parse_methods -- parse_pos_cas -- parse_pos_model_input)) >>
      (fn (name, (id, ((((source, maa), mets), cas), model_input))) =>
        Toplevel.theory (fn thy =>
          let
            val ctxt = Proof_Context.init_global thy
            val pbl_id = References_Def.explode_id id;
            val met_ids = map References_Def.explode_id mets;
            val (model, where_) = Model_Pattern.parse_pos ctxt model_input
            val (model', where_') = (map drop_pos_model model, map drop_pos_where_ where_)
            val cas = CAS_Def.parse ctxt cas
            val set_data =
              ML_Context.expression (Input.pos_of source)
                (ml "Theory.setup (Problem.set_data (" @ ML_Lex.read_source source @ ml "))")
              |> Context.theory_map;
            val where_rls = the_data (set_data thy);
            val arg = prep_store thy name maa id_empty
              (model', where_', cas) (pbl_id, met_ids, where_rls)
         in Know_Store.add_pbls ctxt [arg] thy end)))

in end;



(** get Problem from Store **)

fun adapt_to_type ctxt ({guh, mathauthors, start_refine, thy, cas, solve_mets, where_rls, where_, model} : Probl_Def.T) =
  let
    val cas = case cas of
          SOME t => SOME (ParseC.adapt_term_to_type ctxt t)
        | NONE => NONE
    val where_ = map (ParseC.adapt_term_to_type ctxt) where_
    val model = map (Model_Pattern.adapt_to_type ctxt) model
  in
    {guh = guh, mathauthors = mathauthors, start_refine = start_refine, thy = thy, cas = cas, solve_mets = solve_mets,
      where_rls = where_rls, where_ = where_, model = model}
    end

fun from_store ctxt id =
  let
    val pbl = Store.get (get_pbls ()) id (rev id)
  in adapt_to_type ctxt pbl end
    handle ERROR _ =>
      raise error ("*** ERROR Problem.from_store " ^ strs2str id ^ " not found");

(**)end(*struct*)
